%h1 Test-Driven Development in C

%p
  Writing a test harness for programs written in C requires some scaffolding,
  but I agree with Theo de Raadt is correct when he observed that
  = succeed '.' do
    %a{:href=>"http://marc.info/?l=openbsd-ports&m=139474670315494"}<
      heavy frameworks get in the way
%p
  These are some techniques for building a test runner which is capable of
  mocking individual function calls.

%h2 A Simple Test Runner

%p
  Start by including the header file of the module or system under test.  The
  following is based on
  = succeed ',' do
    %a{:href=>"http://www.jera.com/techinfo/jtns/jtn002.html"} MinUnit
  a minimal unit testing framework for C by John Brewer

:codeblock
  :::c
  /* runner.c */

  #include <stdio.h>
  #include "sut.h"

  int tests_run = 0;

%p
  Next define some simple macros that will print a meaningful error and return
  %code 1
  if the test condition is not true

:codeblock
  :::c
  #define FAIL() printf("\nfailure in %s() line %d\n", __func__, __LINE__)
  #define _assert(test) do { if (!(test)) { FAIL(); return 1; } } while(0)
  #define _verify(test) do { int r=test(); tests_run++; if(r) return r; } while(0)

%p
  %code do { ... } while(0)
  is a convention for including multiple statements in a macro; the block
  gives variable declarations local storage.
%p
  The test runner itself is responsible for calling the
  %code run_test
  macro, run each test in turn until one fails.

:codeblock
  :::c
  int square_01() {
      int x=5;
      _assert(square(x) == 25);
      return 0;
  }

  int all_tests() {
      _verify(square_01);
      return 0;
  }

  int main(int argc, char **argv) {
      int result = all_tests();
      if (result == 0)
          printf("PASSED\n");
      printf("Tests run: %d\n", tests_run);

      return result != 0;

%h2 Link-level Stubbing

%p
  Among the techniques available for faking functions in C, link-level stubs are
  a great place to begin because they require no changes to production code.
  One of the first functions to replace is
  %code main()
  so that the test runner is the new entry point for the resulting binary.

:codeblock
  :::make
  runner: runner.o sut.o
  →   strip -N main sut.o _sut.o
  →   ${CC} -o $@ runner.o _sut.o ${LDFLAGS}

%p
  Here we create a stripped version of the object code
  %code utility.o
  which allows the linker to connect the new main from
  = succeed '.' do
    %code runner.o
  Ideally tesrunner is executed as part of the build.

:codeblock
  :::make
  all: utility runner test

  test:
  →   ./runner

/
  Note: Linking a test runner only works if functions have a global scope

%h2 Faking Interfaces with the C Preprocessor

%p
  The most straitfoward way to inject test code into each module is with the C
  preprocessor. This is a very flexible technique, although it may require
  separate makefile to generate the test build.

%p
  Everything needs to be built twice, so it really helps to use a fast
  compiler. Build scripts and makefiles are making very strategic substitutions
  here, so I would strongly recommend staying clear of auto-tools. Instead
  factor out common elements and include them in platform-specific build
  configurations.

:codeblock
  :::make
  sut.o:
  →   @echo CC $<
  →   @${CC} -c -DTEST ${CFLAGS} $<

%p
  Preprocessor instructions can be placed anywhere, but one way of organizing
  code is to break it into two distinct segments. The first is the system under
  test, which contain all of the core functionality of the module, the second
  will include
  %code main()
  and other definitions that may be replaced by fake functions in the
  specification header.

:codeblock
  :::c
  /* functions to be tested */

  #ifdef TEST
  #include "sut_spec.h"
  #else

  /* functions to be faked */

  #endif

%h2 Global Function Pointers

%p
  Functions can be aliased with a function pointer which the test runner can
  then redefine at runtime.  This is the approach I took when writing unit tests
  for
  %a{:href=>"http://github.com/eradman/entr"} entr

:codeblock
  :::c
  #include <stdlib.h>

  /* function pointers */
  char * (*xrealpath)(const char *, char *);

  int main() {
      xrealpath = realpath;
  }

%p
  Instead of assigning the real function, the test runner can stub the result

:codeblock
  :::c
  #include "main.c"

  char *
  fake_realpath(const char *pathname, char *resolved) {
      snprintf(resolved, PATH_MAX, "/home/user/%s", pathname);
      return resolved;
  }

  int test_main(int argc, char *argv[]) {
      xrealpath = fake_realpath;
  }

%h2 Toy Programs

%p
  Ten years after writing this post on techniques for mocking function in C, I
  have adopted a different tact, which is to write small programs that exercise
  parts of a larger library.  This is the pattern employed by the test harness for
  %a{:href=>"http://github.com/eradman/rset"} rset

:codeblock
  :::c
  #include <stdio.h>
  #include "rutils.h"

  int main(int argc, char *argv[])
  {
      if (argc != 3) {
          fprintf(stderr, "usage: ./copyfile src dst\n");
          return 1;
      }
      install_if_new(argv[1], argv[2]);

      return 0;
  }

%p
  These utilities can then be run in a temporary environment using an external
  test framework written in a scripting language.
